#include "adlib.h"
#include <lunar/fx.hpp>
#include <lunar/dsp.h>
#include "kemuopl.h"

const unsigned char op_table[9] =
  {0x00, 0x01, 0x02, 0x08, 0x09, 0x0a, 0x10, 0x11, 0x12};

// Standard AdLib frequency table
const int fnums[] = { 0x16b,0x181,0x198,0x1b0,0x1ca,0x1e5,0x202,0x220,0x241,0x263,0x287,0x2ae };  

class adlib : public lunar::fx<adlib> {
public:
	unsigned char regs[0x100];
	CKemuopl* opl;

	adlib() {
		opl = 0;
	}

	void init() {
		opl = new CKemuopl(44100, true, true);
		memset(regs, 0, sizeof(regs));
 	}
	
	void exit() {
		delete opl;
		delete this;
	}
	
	void setbit(unsigned char& op, int i, bool enabled) {
		if (enabled)
			op |= (1 << i); else
			op = op & (0xFF ^ (1 << i));
		
		printf("op is now %i\n", op);
	}

	void opl_write(int reg) {
		opl->write(reg, regs[reg]);
	}
	
	void process_events() {
		for (int t = 0; t < track_count; ++t) {
			unsigned char op = op_table[t];
			bool reg20_changed = false;
			bool reg23_changed = false;
			bool reg40_changed = false;
			bool reg43_changed = false;
			bool reg60_changed = false;
			bool reg63_changed = false;
			bool reg80_changed = false;
			bool reg83_changed = false;
			bool regbd_changed = false;
			bool regc0_changed = false;
			bool rege0_changed = false;
			bool rege3_changed = false;
			
			// FM operator 1:
			if (tracks[t].ampmod1) {
				// set or clear bit 7
				setbit(regs[0x20 + op], 7, *tracks[t].ampmod1);
				reg20_changed = true;
			}
			if (tracks[t].vib1) {
				setbit(regs[0x20 + op], 6, *tracks[t].vib1);
				reg20_changed = true;
			}
			if (tracks[t].kbscale1) {
				setbit(regs[0x20 + op], 4, *tracks[t].kbscale1);
				reg20_changed = true;
			}
			if (tracks[t].harmonic1) {
				// bits 3-0 of register 20 specifies harmonic scale
				unsigned char& o = regs[0x20 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].harmonic1;
				reg20_changed = true;
			}
			if (tracks[t].lvlkeyscaling1) {
				// bits 7-6 of register 40 specifies key scaling
				unsigned char& o = regs[0x40 + op];
				o = (o & 0x3f) | ((unsigned char)*tracks[t].lvlkeyscaling1 << 6);
				reg40_changed = true;
			}
			if (tracks[t].outputlevel1) {
				// bits 5-0 of register 40 specifies total output level
				unsigned char& o = regs[0x40 + op];
				o = (o & 0xc0) | (unsigned char)*tracks[t].outputlevel1;
				reg40_changed = true;
			}
			if (tracks[t].attack1) {
				// bits 7-4 of register 60 specifies attack
				unsigned char& o = regs[0x60 + op];
				o = (o & 0x0F) | ((unsigned char)*tracks[t].attack1 << 4);
				reg60_changed = true;
			}
			if (tracks[t].decay1) {
				// bits 3-0 of register 60 specifies decay
				unsigned char& o = regs[0x60 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].decay1;
				reg60_changed = true;
			}
			if (tracks[t].sustain1) {
				// bits 7-4 of register 80 specifies systain
				// bit 5 of register 20 enables sustain
				unsigned char sus = *tracks[t].sustain1;
				if (sus == 0) {
					setbit(regs[0x20 + op], 5, false);
					reg20_changed = true;
				} else {
					setbit(regs[0x20 + op], 5, true);
					unsigned char& o = regs[0x80 + op];
					o = (o & 0x0F) | ((sus - 1) << 4);

					reg20_changed = true;
					reg80_changed = true;
				}
			}
			if (tracks[t].release1) {
				// bits 3-0 of register 80 specifies release
				unsigned char& o = regs[0x80 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].release1;
				reg80_changed = true;
			}
			if (tracks[t].wave1) {
				// bits 1-0 of register e0 specifies release
				unsigned char& o = regs[0xe0 + op];
				o = (unsigned char)*tracks[t].wave1;
				rege0_changed = true;
			}
			
			
			
			// FM operator 2:
			if (tracks[t].ampmod2) {
				// set or clear bit 7
				setbit(regs[0x23 + op], 7, *tracks[t].ampmod2);
				reg23_changed = true;
			}
			if (tracks[t].vib2) {
				setbit(regs[0x23 + op], 6, *tracks[t].vib2);
				reg23_changed = true;
			}
			if (tracks[t].kbscale2) {
				setbit(regs[0x23 + op], 4, *tracks[t].kbscale2);
				reg23_changed = true;
			}
			if (tracks[t].harmonic2) {
				// bits 3-0 of register 20 specifies harmonic scale
				unsigned char& o = regs[0x23 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].harmonic2;
				reg23_changed = true;
			}
			if (tracks[t].lvlkeyscaling2) {
				// bits 7-6 of register 40 specifies key scaling
				unsigned char& o = regs[0x43 + op];
				o = (o & 0x3f) | ((unsigned char)*tracks[t].lvlkeyscaling2 << 6);
				reg43_changed = true;
			}
			if (tracks[t].outputlevel2) {
				// bits 5-0 of register 40 specifies total output level
				unsigned char& o = regs[0x43 + op];
				o = (o & 0xc0) | (unsigned char)*tracks[t].outputlevel2;
				reg43_changed = true;
			}
			if (tracks[t].attack2) {
				// bits 7-4 of register 60 specifies attack
				unsigned char& o = regs[0x63 + op];
				o = (o & 0x0F) | ((unsigned char)*tracks[t].attack2 << 4);
				reg63_changed = true;
			}
			if (tracks[t].decay2) {
				// bits 3-0 of register 60 specifies decay
				unsigned char& o = regs[0x63 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].decay2;
				reg63_changed = true;
			}
			if (tracks[t].sustain2) {
				// bits 7-4 of register 80 specifies systain
				// bit 5 of register 20 enables sustain
				unsigned char sus = *tracks[t].sustain2;
				if (sus == 0) {
					setbit(regs[0x23 + op], 5, false);
					reg23_changed = true;
				} else {
					setbit(regs[0x23 + op], 5, true);
					unsigned char& o = regs[0x83 + op];
					o = (o & 0x0F) | ((sus - 1) << 4);

					reg23_changed = true;
					reg83_changed = true;
				}
			}
			if (tracks[t].release2) {
				// bits 3-0 of register 80 specifies release
				unsigned char& o = regs[0x83 + op];
				o = (o & 0xF0) | (unsigned char)*tracks[t].release2;
				reg83_changed = true;
			}
			if (tracks[t].wave2) {
				// bits 1-0 of register e0 specifies release
				unsigned char& o = regs[0xe3 + op];
				o = (unsigned char)*tracks[t].wave2;
				rege3_changed = true;
			}
			
			

			if (tracks[t].ampmod) {
				setbit(regs[0xbd + op], 7, *tracks[t].ampmod);
				regbd_changed = true;
			}

			if (tracks[t].vib) {
				setbit(regs[0xbd + op], 6, *tracks[t].vib);
				regbd_changed = true;
			}
			
			if (tracks[t].feedback) {
				// bits 3-1 of register c0 specifies feedback
				unsigned char& o = regs[0xc0 + op];
				o = (unsigned char)*tracks[t].feedback << 1;
				regc0_changed = true;
			}
			if (tracks[t].algo) {
				// bits 0 of register c0 specifies algorithm
				unsigned char& o = regs[0xc0 + op];
				o = (unsigned char)*tracks[t].algo & 0x1;
				regc0_changed = true;
			}
		
			if (reg20_changed)
				opl_write(0x20 + op);
			if (reg23_changed)
				opl_write(0x23 + op);
			if (reg40_changed)
				opl_write(0x40 + op);
			if (reg43_changed)
				opl_write(0x43 + op);
			if (reg60_changed)
				opl_write(0x60 + op);
			if (reg63_changed)
				opl_write(0x63 + op);
			if (reg80_changed)
				opl_write(0x80 + op);
			if (reg83_changed)
				opl_write(0x83 + op);
			if (regbd_changed)
				opl_write(0xbd + op);
			if (regc0_changed)
				opl_write(0xc0 + op);
			if (rege0_changed)
				opl_write(0xe0 + op);
			if (rege3_changed)
				opl_write(0xe3 + op);
			

			
			/*if (tracks[t].para20) opl->write(0x20 + op, *tracks[t].para20);
			if (tracks[t].para23) opl->write(0x23 + op, *tracks[t].para23);
			if (tracks[t].para40) opl->write(0x40 + op, *tracks[t].para40);
			if (tracks[t].para43) opl->write(0x43 + op, *tracks[t].para43);
			if (tracks[t].para60) opl->write(0x60 + op, *tracks[t].para60);
			if (tracks[t].para63) opl->write(0x63 + op, *tracks[t].para63);
			if (tracks[t].para80) opl->write(0x80 + op, *tracks[t].para80);
			if (tracks[t].para83) opl->write(0x83 + op, *tracks[t].para83);
			if (tracks[t].parae0) opl->write(0xe0 + op, *tracks[t].parae0);
			if (tracks[t].parae3) opl->write(0xe3 + op, *tracks[t].parae3);
			if (tracks[t].parac0) opl->write(0xc0 + op, *tracks[t].parac0);*/

			if (tracks[t].note) {
				
				if (*tracks[t].note == 0) {
					printf("adlib note off!\n");
					regs[0xb0 + t] &= 0xdf; // This means the "Key On" bit will always be 0
					opl_write(0xb0 + t);
					continue;
				}

				int note = (int)*tracks[t].note;
				int freq = fnums[note%12];
				int oct = note/12;
				regs[0xa0 + t] = (unsigned char)(freq&0xff);
				int c=((freq&0x300) >> 8)+(oct<<2) + (1<<5);
				regs[0xb0 + t] = (unsigned char)c;
				opl_write(0xa0 + t);
				opl_write(0xb0 + t);
			}
	      }
      }
	
	void process_stereo(float *inL, float *inR, float *outL, float *outR, int n) {
		short buffer[256 * 2];
		opl->update(buffer, n);
		for (int i = 0; i < n; i++) {
			*outL++ = (float)buffer[i * 2 + 0] / 32767.0f;
			*outR++ = (float)buffer[i * 2 + 1]  / 32767.0f;
		}
		
	}
};

lunar_fx *new_fx() {
  return new adlib();
}
