/*
 * ADLIBEMU.H
 * Copyright (C) 1998-2001 Ken Silverman
 * Ken Silverman's official web site: "http://www.advsys.net/ken"
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#pragma once

#if !defined(max) && !defined(__cplusplus)
#define max(a,b)  (((a) > (b)) ? (a) : (b))
#endif
#if !defined(min) && !defined(__cplusplus)
#define min(a,b)  (((a) < (b)) ? (a) : (b))
#endif

#define PI 3.141592653589793
#define MAXCELLS 18
#define WAVPREC 2048

#define FRQSCALE (49716/512.0)

//Constants for Ken's Awe32, on a PII-266 (Ken says: Use these for KSM's!)
#define MODFACTOR 4.0      //How much of modulator cell goes into carrier
#define MFBFACTOR 1.0      //How much feedback goes back into modulator
#define ADJUSTSPEED 0.75   //0<=x<=1  Simulate finite rate of change of state

//Constants for Ken's Awe64G, on a P-133
//#define MODFACTOR 4.25   //How much of modulator cell goes into carrier
//#define MFBFACTOR 0.5    //How much feedback goes back into modulator
//#define ADJUSTSPEED 0.85 //0<=x<=1  Simulate finite rate of change of state

#define FIFOSIZ 256


struct celltype
{
    float val, t, tinc, vol, sustain, amp, mfb;
    float a0, a1, a2, a3, decaymul, releasemul;
    short *waveform;
    long wavemask;
    void (*cellfunc)(void *, float);
    unsigned char flags, dum0, dum1, dum2;
};

struct adlibemu {

	
	long numspeakers, bytespersample;
	float recipsamp;
	celltype cell[MAXCELLS];
	signed short wavtable[WAVPREC*3];

	unsigned char adlibreg[256], ksl[8][16];
	unsigned char odrumstat;

	float lvol[9],rvol[9];
	long lplc[9],rplc[9];
	
	float AMPSCALE;
	float *rptr[9], *nrptr[9];
	float rbuf[9][FIFOSIZ*2];
	float snd[FIFOSIZ*2];
	long nlvol[9], nrvol[9];
	long nlplc[9], nrplc[9];
	long initfirstime;
	long rend;

	adlibemu();
	
	void adlibinit(long dasamplerate,long danumspeakers,long dabytespersample);
	void adlib0(long i,long v);
	void adlibgetsample(void *sndptr,long numbytes);
	void adlibsetvolume(int i);
	void randoinsts();

	
	void cellon (long i, long j, celltype *c, unsigned char iscarrier);
	void cellfreq (signed long i, signed long j, celltype *c);

};
